//
// canola - canon canola 1614p emulator
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef CANOLA_NIXIE_TUBES_H
#define CANOLA_NIXIE_TUBES_H

#include <lib/config.h>
#include <gtkmm/box.h>
#include <gtkmm/eventbox.h>

#include <lib/display.h>

#include <canola/nixie_tube.h>

/**
  * The nixie_tubes class is used to represent the array of 16 nixie
  * tubes on the calculator.
  */
class nixie_tubes:
    public Gtk::VBox
{
public:
    /**
      * The destructor.
      */
    virtual ~nixie_tubes();

    /**
      * The default constructor.
      */
    nixie_tubes();

    /**
      * The set_value method may be used to set the appearance of the
      * nixie tubes and indicator lamps, whenever the indicator display
      * may have changed.
      *
      * @param value
      *     The value to be displayed.
      */
    void set_value(const display &value);

private:
    /**
      * The overflow_lamp instance variable is used to remember appearance
      * and state of the Overflow Lamp to the left of the nixie tubes.
      *
      *
      * "Overflow Lamp
      *
      * "Glows in any one of the following cases, causing entry and operation
      * to be stopped.
      * 1) When entry exceeds 16 digits.
      * 2) When two or more keys are touched simultaneously by mistake.
      * 3) When the integer part of the calculation result exceeds 16 digits.
      * 4) When the integer part of the value stored in a memory exceeds
      *    16 digits.
      */
    Gtk::Image overflow;

    /**
      * The bezal_left instance variable is used to remember the
      * appearance of the filler widget for the left hand side of the
      * recess the nixie tubes sit in.
      */
    Gtk::Image bezel_left;

    /**
      * The bnixie_tubes instance variable is used to remember the
      * appearance and state of each of the nixie tubes.  They include a
      * little of the bezel, as well.
      */
    nixie_tube tubes[16];

    /**
      * The bezal_right instance variable is used to remember the
      * appearance of the filler widget for the right hand side of the
      * recess the nixie tubes sit in.
      */
    Gtk::Image bezel_right;

    /**
      * The minus_lamp instance variable is used to remember appearance
      * and state of the Minus Lamp to the right of the nixie tubes.
      *
      *
      * "Minus Lamp
      *
      * "Glows when the value is negative, indicating that the indicated value
      * is negative."
      */
    Gtk::Image negative;

    /**
      * The tube instance variable is used to remember the positions and
      * scales of the overflow indicator, nixie tubes and negativive
      * indicator.
      */
    Gtk::HBox tube_layout;

    /**
      * The tube_bg instance variable is used to remember the background
      * color for the nixie tubes and upper indicator lamps.
      */
    Gtk::EventBox tube_bg;

    /**
      * The m1_lamp instance variable is used to remember appearance and
      * state of the M1 lamp.
      *
      *
      * "Memory 1 Lamp
      *
      * "Glows when values other than zero are stored in Memory 1, indicating
      * that Memory 1 is in use.  Goes off automatically when Memory 1 is
      * cleared by the [CM1] key, indicating that no values are stored in
      * Memory 1."
      */
    Gtk::Image m1_lamp;

    /**
      * The m2_lamp instance variable is used to remember the appearance
      * and state of the M2 lamp.
      *
      *
      * "Memory 2 Lamp
      *
      * "Glows when values other than zero are stored in Memory 2, indicating
      * that Memory 2 is in use.  Goes off automatically when Memory 2 is
      * cleared by the [CM2] key, indicating that no values are stored in
      * Memory 2."
      */
    Gtk::Image m2_lamp;

    /**
      * The entry_lamp instance variable is used to remember
      * the appearance and state of the Entry Lamp.
      *
      *
      * "Entry Lamp
      *
      * "This lamp is used when executing a program.  The lamp glows when
      * program advancement is stopped by ENT, SJ or EJ instructions,
      * indicating that the cakculator is waiting for instruction from the
      * operator.  The four instructions to be put in by the operator are as
      * follows.
      *
      * (i)   The program is advanced as it is by the [START] key operation.
      *       (ENT, SJ, EJ)
      *
      * (ii)
      *       Data is read in by the operation of entry keys and the [START]
      *       key, and the program calculation is advanced to the following
      *       step.  (ENT)
      *
      * (iii) By the operation of the [SJ] key, the program is caused to jump
      *       to the specified step. (SJ)
      *
      *       If the [STSRT] key is depressed instead, the program proceeds
      *       without jumping.  (SJ)
      *
      * (iv)  While data is being read into the calculator by the operation of
      *       entry keys and the [START] key, the program is caused to jump to
      *       the specified step.
      *
      *       If the [START] key is depressed without entry, the program
      *       proceeds without jumping.
      */
    Gtk::Image entry_lamp;

    /**
      * The lamp_layour instance variable is used to remember the
      * positions of the three indicator lamps below the nixie tubes.
      */
    Gtk::HBox lamp_layout;

    /**
      * The lamp_bg instance variable is used to remember the background
      * color for the three indicator lamps below the nixie tubes.
      */
    Gtk::EventBox lamp_bg;

    /**
      * The images instance variable is used to remember the varous
      * images required to rember the various indicator lamps.
      */
    Glib::RefPtr<Gdk::Pixbuf> images[14];

    /**
      * The load_images method is used to load the images required to
      * display the nixie tubes and other indicator lamps.
      */
    void load_images(void);

    /**
      * The copy constructor.
      * Do not use.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    nixie_tubes(const nixie_tubes &rhs);

    /**
      * The assignment operator.
      * Do not use.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    nixie_tubes &operator=(const nixie_tubes &rhs);
};

#endif // CANOLA_NIXIE_TUBES_H
