//
// canola - canon canola 1614p emulator
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef CANOLA_RETRO_BUTTON_H
#define CANOLA_RETRO_BUTTON_H

#include <lib/config.h>
#include <gtkmm.h>

/**
  * The retro_button class is used to represent the processing required
  * to draw a retro button in the style of the original calculator.
  */
class retro_button:
    public Gtk::EventBox
{
public:
    /**
      * The destructor.
      */
    virtual ~retro_button();

    /**
      * The constructor.
      *
      * @param caption
      *     The label to put on the top of the button.
      * @param color
      *     The vague color name of the button.
      * @param text_fg
      *     The foreground color to use for the caption.
      * @param toggle
      *     whether or not the button is a toggle button (true) or a
      *     normal button (false).
      */
    retro_button(const char *caption, const char *color,
        const Gdk::RGBA &text_fg, bool toggle = false);

    /**
      * The signal_clicked method is used to obtain the signal that will
      * be issued when the button is clicked.
      */
    sigc::signal<void> signal_clicked(void) { return signal_clicked_sender; }

    /**
      * The signal_released method is used to obtain the signal that will
      * be issued when the button is released (toggle buttons only).
      */
    sigc::signal<void> signal_released(void) { return signal_released_sender; }

private:
    /**
      * The normal_image instance variable is used to remember when the
      * button looks like when it has not been pressed.
      */
    Glib::RefPtr<Gdk::Pixbuf> normal_image;

    /**
      * The normal_image instance variable is used to remember when the
      * button looks like when it has been pressed.
      */
    Glib::RefPtr<Gdk::Pixbuf> active_image;

    /**
      * The normal_image instance variable is used to remember when the
      * button looks like when it has been disabled.
      */
    Glib::RefPtr<Gdk::Pixbuf> disabled_image;

    /**
      * The appearance instance variable is used to remember the state
      * of the image widget displaying the image that looks like the
      * button in it's present state.
      */
    Gtk::Image appearance;

    /**
      * The text instance variable is used to remember the state of
      * widget displaying the button's label.
      */
    Gtk::Label text;

    /**
      * The text_aligner instance variable is used to remember the state
      * of the widget keeping the label centered over the present state
      * image.
      */
    Gtk::Alignment text_aligner;

    /**
      * The layout instance variable is used to remember the state of
      * the way the image and the label are displayed for the button.
      * In particular, the label is over the top of the button, but one
      * pixel offset when pressed.
      */
    Gtk::Fixed layout;

    /**
      * The signal_clicked_sender instance variable is used to remember
      * the callbacks subscribed to the "clicked" signal.
      */
    sigc::signal<void> signal_clicked_sender;

    /**
      * The signal_released_sender instance variable is used to remember
      * the callbacks subscribed to the "released" signal.
      */
    sigc::signal<void> signal_released_sender;

    /**
      * The inside instance variable is used to remember whether or not
      * the pointer is inside the button at the moment.
      */
    bool inside;

    /**
      * The pressed instance variable is used to remember whether or not
      * the button is pressed at the moment.
      */
    bool pressed;

    /**
      * The toggle instance variable is used to remember whether or no
      * this button is a toggle button.
      */
    bool toggle;

    /**
      * The pressed_callback method is used to handle the pressed signal
      * passed to us by the parent (base) class pressed signal.  The
      * constructor subscribes to the signal.
      */
    bool pressed_callback(const _GdkEventButton *event);

    /**
      * The released method is used to handle the released signal
      * passed to us by the parent (base) class released signal.  The
      * constructor subscribes to the signal.
      */
    bool released_callback(const _GdkEventButton *event);

    /**
      * The enter_callback method is used to handle the enter signal
      * passed to us by the parent (base) class released signal.  The
      * constructor subscribes to the signal.
      */
    bool enter_callback(const _GdkEventCrossing *event);

    /**
      * The fubar method is used to to handle the leave signal passed to
      * us by the parent (base) class released signal.  The constructor
      * subscribes to the signal.
      */
    bool leave_callback(const _GdkEventCrossing *event);

    /**
      * The up method is used to perform the processing attendant to the
      * button going up.  It switches images and moves the label slightly.
      */
    void up(void);

    /**
      * The down method is used to perform the processing attendant to
      * the button going down.  It switches images and moves the label
      * slightly.
      */
    void down(void);

    /**
      * The active instance variable is used to remember whether or not
      * this button is in the "active" state at present.
      */
    bool active;

    /**
      * The default constructor.  Do not use.
      */
    retro_button();

    /**
      * The copy constructor.  Do not use.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    retro_button(const retro_button &rhs);

    /**
      * The assignment operator.  Do not use.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    retro_button &operator=(const retro_button &rhs);
};

#endif // CANOLA_RETRO_BUTTON_H
