//
// canola - canon canola 1614p emulator
// Copyright (C) 2012 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef CANOLA_SLIDER_H
#define CANOLA_SLIDER_H

#include <lib/config.h>
#include <gtkmm.h>
#include <vector>

/**
  * The slider class is used to represent a control that slides
  * horizontally to assume one of several positions.
  */
class slider:
    public Gtk::EventBox
{
public:
    /**
      * The destructor.
      */
    virtual ~slider();

    typedef sigc::signal<void> value_changed_signal_t;

    /**
      * The value_changed_signal method is used to obtain access to
      * the signal that will be emitted each time the user changes the
      * switch value.
      *
      * @note
      *     The set_value method will <b>not</b> cause this signal to be
      *     emitted.
      */
    value_changed_signal_t &value_changed_signal(void);

    /**
      * The get_value method may be used to obtain the current value of
      * the widget.
      */
    int get_value(void) const;

    /**
      * The set_value method is used to set the value of the switch.
      *
      * @param n
      *     The new value of the widget (0..<nvalues).
      *     Out-of-range values will be clipped.
      */
    void set_value(int n);

protected:
    /**
      * The constructor.
      * For use by derived classes only.
      *
      * @param name
      *     The base name of the images used to display the slider.
      * @param npos
      *     The number of positions (0..<n) the slider can assume.
      * @param act_on_rel
      *     Whether the widget is activated on release (true), or always
      *     active.
      */
    slider(const std::string &name, int npos, bool act_on_rel = false);

private:
    /**
      * The nvalues instance variable is used to remember the number of
      * valid values.
      *
      * assert(nvalues >= 2);
      */
    int nvalues;

    /**
      * The value instance variable is used to remember the current
      * value of the slider.
      *
      * assert(value >= 0);
      * assert(value < nvalues);
      */
    int value;

    typedef Glib::RefPtr<Gdk::Pixbuf> pixbuf_pointer;

    /**
      * The value_images instance variable is used to remember the
      * pixbuf corresponding to each value.
      */
    std::vector<pixbuf_pointer> value_images;

    /**
      * The appearance instance variable is used to remember the state
      * of the image widget displaying the pixbuf that looks like the
      * slider in it's present state.
      */
    Gtk::Image appearance;

    /**
      * The down instance variable is used to remember whether or not
      * the user is currently clicking on this switch.
      */
    bool down;

    /**
      * The inside instance variable is used to remember whether the
      * pointer is currently inside or outside the widget.  Only
      * meaningful when #down is active.
      */
    bool inside;

    /**
      * The old_value instance variable is used to remember what the
      * slider value was immediately before the slider was clicked.
      * This value will be restored whenever the pointer goes outside
      * the widget.
      */
    int old_value;

    /**
      * The frac instance variable is used to remember how wide each
      * value setting is.  This is the width of the images divided by
      * the #nvalues.
      */
    int frac;

    /**
      * The on_button_press method is used to handle button click events
      * generated by the user.
      */
    bool on_button_press(GdkEventButton *ep);

    /**
      * The on_button_release method is used to handle button unclick
      * events generated by the user.
      */
    bool on_button_release(GdkEventButton *ep);

    /**
      * The on_motion_notify method is used to handle motion events
      * generated by the user.
      */
    bool on_motion_notify(GdkEventMotion *ep);

    /**
      * The update_value method is used to take the position indivcated
      * by an event, and determine whether or not the switch value
      * has changed.  If it has, the image will change, and a
      * #value_changed_source signal will be emitted.
      */
    void update_value(int x, int y);

    /**
      * The value_changed_source instance variable is used to remember
      */
    value_changed_signal_t value_changed_source;

    /**
      * The activate_on_release
      * insta
      */
    bool activate_on_release;

    /**
      * The default constructor.
      * Do not use.
      */
    slider();

    /**
      * The copy constructor.
      * Do not use.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    slider(const slider &rhs);

    /**
      * The assignment operator.
      * Do not use.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    slider &operator=(const slider &rhs);
};

#endif // CANOLA_SLIDER_H
