//
// canola - canon canola 1614p emulator
// Copyright (C) 2012 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef CANOLA_WINDOW_CARD_READER_H
#define CANOLA_WINDOW_CARD_READER_H

#include <lib/config.h>
#include <gtkmm.h>
#include <vector>
#include <boost/shared_ptr.hpp>

#include <canola/calculator/proxy.h>

class calculator_gnome;

/**
  * The window_card_reader class is used to represent a window within
  * which the card images are animated to simulate a card reader drawing
  * each card in and passing the holes over the optical sensors.
  */
class window_card_reader:
    public Gtk::Window
{
public:
    typedef Glib::RefPtr<window_card_reader> pointer;

    /**
      * The destructor.
      */
    virtual ~window_card_reader();

    /**
      * The create class method is used to create new dynamicall
      * allocated isnatnces of this class.
      *
      * @param dst
      *     The destination, the calulator into which the program is to
      *     be loaded.
      */
    static pointer create(calculator_gnome *dst);

    /**
      * The start method is used to start the card reader animation going.
      *
      * @param src
      *     The source, the file to read to obtain the program.
      * @param binary
      *     Whether or not he file containing the program to be loaded
      *     is assembler text (false) or binary (true).
      * @returns
      *     true if the program was read successfully, false if there
      *     was one or more errors (erorrs reported via dst.on_error
      *     already).
      */
    bool start(const std::string &src, bool binary = false);

    /**
      * The is_busy method may be used to determine whether or not the
      * card reader is currently animating the reading of a card set.
      */
    bool is_busy();

private:
    /**
      * The constructor.
      * It is private on purpose, use the #create class method instead.
      */
    window_card_reader(calculator_gnome *dst);

    /**
      * The dst instance variable is used to remember the program, and
      * also where to forward opcodes to, at the right time.
      */
    calculator_proxy dst;

    /**
      * The layout instance variable is used to remember the positions
      * of each of the card image widgets.
      */
    Gtk::Fixed layout;

    typedef Glib::RefPtr<Gdk::Pixbuf> pixbuf_ptr_t;

    typedef std::vector<pixbuf_ptr_t> card_pixbufs_t;

    /**
      * The card_pixbufs instance variable is used to remember the
      * images of each vcard to be animated.  They are actually faithful
      * images of the cards that the program would require.
      */
    card_pixbufs_t card_pixbufs;

    typedef boost::shared_ptr<Gtk::Image> gtk_image_ptr_t;

    typedef std::vector<gtk_image_ptr_t> card_image_widgets_t;

    /**
      * The card_image_widgets instance variable is used to remember
      * the Gtk::Image instances, one for each card pixbuf.
      */
    card_image_widgets_t card_image_widgets;

    /**
      * The empty method is used to determine whether or not any cards
      * are presently loaded into the card reader simulation.
      */
    bool empty(void) const;

    /**
      * The animation_timer instance variable is used to remember
      * the timer being used to animate the cards, vis a the
      * #on_animation_timer method.
      */
    sigc::connection animation_timer;

    /**
      * The pixels_per_second instance variable is used to remember
      * how man pixels of the card image pass under the read head per
      * second.  This ensures that each card takes 4 seconds to read,
      * but this isn't a constant becaise each display is different.
      */
    double pixels_per_second;

    /**
      * The pixels_per_instruction instance variable is used to
      * remember how many pixels pass under the read position for
      * one instruction.  This is used within #on_animation_timer to
      * determine when to forward opcodes.
      */
    int pixels_per_instruction;

    /**
      * The current_address instance variable is used to remember the
      * address from which to fetch the next opcode to be forwarded to
      * the #dst destination.
      */
    int current_address;

    /**
      * The current_card_number instance variable is used to remember
      * which card is currently being animated.  Once this exceeds the
      * size of #card_image_widgets, we are done.
      */
    size_t current_card_number;

    /**
      * The currently_moving_left instance variable is used to remember
      * whether the current card is moving left (true) or moving up
      * (false).  Only when moving up does the card pass under the
      * simulated card reader.
      */
    bool currently_moving_left;

    /**
      * The current_card_started instance variable is used to remember
      * the time in seconds (see getdaytime(2)) that the current phase
      * of the current card started reading.  The time implies the
      * position, thanks to #pixels_per_second.
      */
    double current_card_started;

    /**
      * The on_animation_timer method is used to update the card
      * positions each time the animation timer ticks over.
      */
    bool on_animation_timer(void);

    /**
      * The done method is used to dismiss the window when the last of
      * the last simulated card has passed under the simulated card reader.
      */
    void done(void);

    /**
      * The default constructor.
      * Do not use.
      */
    window_card_reader();

    /**
      * The copy constructor.
      * Do not use.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    window_card_reader(const window_card_reader &rhs);

    /**
      * The assignment operator.
      * Do not use.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    window_card_reader &operator=(const window_card_reader &rhs);
};

#endif // CANOLA_WINDOW_CARD_READER_H
