//
// canola - canon canola 1614p emulator
// Copyright (C) 2011, 2012 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIB_CALCULATOR_H
#define LIB_CALCULATOR_H

#include <lib/ac/stdarg.h>
#include <boost/shared_ptr.hpp>
#include <list>
#include <map>
#include <string>

#include <lib/display.h>
#include <lib/format_printf.h>
#include <lib/interpose.h>
#include <lib/location.h>
#include <lib/number.h>
#include <lib/opcode.h>
#include <lib/precision_and_rounding.h>
#include <lib/sizeof.h>

class debugger; // forward
class program_card_functor; // forward

/**
  * The calculator class is used to represent the interface to a general
  * state of the 1614p emulation.  It isn't necessarily a Gnome widget.
  */
class calculator
{
public:
    typedef boost::shared_ptr<calculator> pointer;

    /**
      * The destructor.
      */
    virtual ~calculator();

    /**
      * The on_opcode method is used to process a key that corresponds
      * to an opcode.  This is most of the keys on the keyboard, but
      * none of the controls above the keys.
      *
      * The actual interpretation of the key depends on several
      * different state paramaters withing the calculator, including:
      * operation mode (LRN, OPE, etc) whether there is a tilde key
      * pending, etc.
      *
      * @param op
      *     The opcode to be processed
      * @param where
      *     The source code location of the opcode (or NULL if not available)
      */
    virtual void on_opcode(opcode_t op, const location::pointer &where);

    /**
      * The on_opcode_lrn method is used by the on_opcode method to
      * process opcodes when in Program Mode LRN.
      *
      * @param op
      *     The opcode to be processed
      * @param where
      *     The source code location of the opcode (or NULL if not available)
      */
    void on_opcode_lrn(opcode_t op, const location::pointer &where);

    /**
      * The on_opcode_ope method is used by the on_opcode method to
      * process opcodes when in Program Mode OPE.
      *
      * @param op
      *     The opcode to be processed
      * @param where
      *     The source code location of the opcode (or NULL if not available)
      */
    void on_opcode_ope(opcode_t op, const location::pointer &where);

    /**
      * The on_opcode_pro_che method is used by the on_opcode method to
      * process opcodes when in Program Mode PRO_CHE.
      *
      * @param op
      *     The opcode to be processed
      * @param where
      *     The source code location of the opcode (or NULL if not available)
      */
    void on_opcode_pro_che(opcode_t op, const location::pointer &where);

    /**
      * The on_opcode_ope_che method is used by the on_opcode method to
      * process opcodes when in Program Mode OPE CHE.
      *
      * @param op
      *     The opcode to be processed
      * @param where
      *     The source code location of the opcode (or NULL if not available)
      */
    void on_opcode_ope_che(opcode_t op, const location::pointer &where);

    /**
      * The on_error method is used to present an error to the user.  The
      * original 1614P did not have this facility, but it sure makes
      * some stuff much easier.
      *
      * The default implementation prints the error on the standard
      * output.  Derived classes may choose to overload it.
      *
      * @param fmt
      *     The format of the error message.  It controls the number
      *     and types of arguments following; see printf(3) for more
      *     information.
      */
    void on_error(const char *fmt, ...)
                                                            FORMAT_PRINTF(2, 3);

    virtual void on_error_v(const char *fmt, va_list ap)
                                                        FORMAT_PRINTF(2, 0) = 0;

    /**
      * The run method is used to run the calculator, from interactive
      * input, via the GUI, or equivalent.
      */
    virtual void run(void) = 0;

    /**
      * The Program Mode Switch controls how the calculator interprets
      * key strokes, either storing a program, or operating as a
      * calculator, or running a program.
      *
      * See the 1614P Instruction Manual, p. 13.
      */
    enum program_mode_t
    {
        /**
          * "1) LRN (Learn)
          *     When storing a program, manually using the keyboard, and when
          *     a program card is read into the calculator or when clearing a
          *     program, the switch is set in this position,"
          */
        program_mode_learn,

        /**
          * "2) OPE (Operation)
          *    (i)  The switch is set in this position when a program
          *         calculation is executed using actual data.
          *    (ii) The switch is also set in this position when an ordinary
          *         manual calculation is made."
          */
        program_mode_operation,

        /**
          * "3) PRO CHE (Program Check)
          *     (i)  The switch is set in this position when checking the
          *          contents of a stored program.
          *     (ii) The switch is also set in this position when partial
          *          correction to the program is to be made."
          */
        program_mode_program_check,

        /**
          * "4) OPE CHE (Operation Check)
          *     Use this position for effecting calculation at each step in
          *     accordance with the contents of the stored program, so that
          *     results may be checked at each step."
          */
        program_mode_operation_check,
    };

    /**
      * The program_mode_get method may be used to obtain the current
      * setting of the calculator's Program Mode Switch.
      */
    program_mode_t program_mode_get(void);

    /**
      * The program_mode_set method may be used to change the setting of
      * the calculator's Program Mode Switch.
      *
      * @param pm
      *     the new program mode
      */
    void program_mode_set(program_mode_t pm);

    /**
      * The Program Selector Switch controls the portion of the
      * calculator's program memory to be used.
      *
      * See the 1614P Instruction Manual, p. 13.
      *
      * "When storing a program, executing a program calculation, and
      * checking or clearing a program, program ranges are selected by
      * this switch"
      */
    enum program_selector_t
    {
        program_selector_i_ii,
        program_selector_i,
        program_selector_ii
    };

    /**
      * The program_selector_get method may be used to obtain the
      * current setting of the calculator's Program Selector Switch.
      */
    program_selector_t program_selector_get(void) const;

    /**
      * The program_selector_set method may be used to change the
      * setting of the calculator's Program Selector Switch.
      *
      * @param ps
      *     the new program selector
      */
    void program_selector_set(program_selector_t ps);

    /**
      * The precision_and_rounding_set method is used to change the
      * precision and rounding in response to the user widgets.
      *
      * @param value
      *     The new value to be set.
      */
    void precision_and_rounding_set(const precision_and_rounding &value);

    /**
      * The precision_and_rounding_get mewthod is used to obtain the
      * current precision and rounding settings, for application to user
      * widgets.
      */
    const precision_and_rounding &precision_and_rounding_get(void) const;

    /**
      * The clear_program method is used to remove the program from the
      * selected bank.
      */
    void clear_program(void);

    /**
      * The load_program method may be used by derived classes to load
      * a program (actually, an opcode stream) from a file.  Errors are
      * reported via the #on_error method.
      *
      * @param filename
      *     The file to load the programme from.
      * @param binary
      *     Whether or not the file is binary (true) or text (false).
      * @returns
      *     true on success, false if there are any errors.
      */
    bool load_program(const std::string &filename, bool binary);

    /**
      * The save_program_as method may be used bu derived classes to
      * save the calculator's stored programme to a file, as binary.
      * Errros are reported via the #on_error method.
      *
      * @param filename
      *     The file to receive the programme.
      * @param binary
      *     Whether or not the file is binary (true) or text (false).
      */
    void save_program_as(const std::string &filename, bool binary);

    /**
      * The accumulate_mode_set method may be used to change the state
      * of the #accumulate_mode instance variable.
      *
      * @param value
      *     The new accumulate mode setting.
      */
    void accumulate_mode_set(bool value);

    /**
      * The consatant_mode_set method is used to select the constant
      * mode setting, to reflect the state of the [K] key.
      *
      * Instruction Manual, p. 10
      *
      * "This key is for constant multiplication and division.  When it
      * is depressed, it locks and function state is attained.
      *
      * "Internal calculation is performed in the normal procedure.
      * Thereafter constant multiplication and division can be made by
      * operating the numeric and [+=] keys.
      *
      * "In the case of multiplication, the multiplicant is a constant,
      * and in the case of division, the divisor is a constant.
      *
      * "This key is also used in involutions.  The "n"th power of a
      * figure is obtain by depressing the [+=] key (n-1) times.  Thus,
      * a [x] [+=] [+=] [+=] [+=] gives a<sup>5</sup>.
      */
    void constant_mode_set(bool value);

    /**
      * The program_card_walk method is used to call the given functor
      * once for each program card in the stored #programme.
      *
      * @param f
      *     The functor to invoke for each card.
      */
    void program_card_walk(program_card_functor &f) const;

    /**
      * The property method is used by the lexer to pass along a
      * property from the program text.
      *
      * @param name
      *     The name of the property.
      * @param value
      *     The value of the property.
      */
    virtual void property(const std::string &name, const std::string &value);

    /**
      * The program_mode_name method may be used to obtain a human
      * readable description of the program mode value.
      *
      * @param pm
      *     The program mode to be described.
      * @returns
      *     a human readable description of the program mode value.
      */
    static const char *program_mode_name(program_mode_t pm);

    /**
      * The strict_memory_number_checking_required method is used to determine
      * whether or not the number keys following M3~ keys (et al) must be
      * followed by a digit key (true) or any key (false).  The actual 1614P
      * silently permits any key, and uses only the lower 4 bits.
      */
    bool strict_memory_number_checking_required(void) const;

    /**
      * The strict_memory_number_checking_set method is used to change
      * the strict memory number checking preference.
      */
    void strict_memory_number_checking_set(bool);

    /**
      * The save_memories_as_program method is used to save the
      * calculators data memories as a program.
      *
      * @param filename
      *     The name of the text file to be written with the assembler
      *     of the program that would reproduce the data memory values.
      * @returns
      *     true on success, false if any error occurred (already
      *     reported via the #on_error method).
      */
    bool save_memories_as_program(const std::string &filename);

protected:
    /**
      * The default constructor.
      * For use by derived classes only.
      */
    calculator();

    /**
      * The update_display method calculates the appropriate display
      * value, and then calls #derived_update_display to cause the
      * derived class to display the approriate value.
      */
    void update_display(void);

    /**
      * The clock_tick_event method is used to inform the claculator
      * that a clock tick has occurred.  This is used to run the
      * program steps, when in Program Mode OPE and the user has
      * pressed the Start button.  Unexpected clock tick events will be
      * discarded.
      *
      * Use the #clock_tick_expected method if you need to know whether
      * or not to re-trigger the clock tick from your callback handler.
      */
    void clock_tick_event(void);

    /**
      * The clock_tick_expected method may be used to determine whether
      * or not the calculator thinks the instruction clock is ticking.
      */
    bool clock_tick_expected(void) const;

    /**
      * The clock_tick_start_running method is called by the calculator
      * to inform the derived class that it should kick off a clock
      * tick idle handler.  This will happen when opcode_extended_start
      * is seen, for example.  Derived classes should not assume when
      * clocks stop and start, instend they will be explicitly informed
      * via this method.
      */
    virtual void clock_tick_start_running(void) = 0;

    /**
      * The printer_finished_event method is used to inform the
      * calculator that the printer is ready to print again.  Unexpected
      * printer fisnished events will be discarded.
      */
    void printer_finished_event(void);

    /**
      * The printer_finished_event_expected method is used to determine
      * whether or not a #printer_finished_event is expected.
      */
    bool printer_finished_event_expected(void) const;

    typedef boost::shared_ptr<debugger> debugger_pointer;

    /**
      * The update_debugger method si used to ask the calculator to pass
      * on its state to the given debugger instance (usually a Gnome,
      * window, but not necessarily).
      *
      * @param where
      *     Where to display the debugging information.
      */
    void update_debugger(debugger &where);

    /**
      * The set_address method is used by derived classes when the user,
      * usually through some sort of GUI, updates the address.
      *
      * @note
      *     This method will NOT call the #update_debugger method.
      */
    void set_address(int n);

    /**
      * The get_step method is used to obtain a particular step of the program.
      *
      * @param addr
      *     The address of interest.
      * @param op
      *     (Output) the opcode at he given location.
      * @param where
      *     (Output) the source file location of the opcode.
      */
    bool get_step(int addr, unsigned char &op, location::pointer &where) const;

private:
    /**
      * The derived_print method is used to implement the PRINT opcode,
      * in the cases where it is valid.
      *
      * @param value
      *     The value to be printed.
      *
      * @note
      *     FIXME: this will not serve, once we start printing program
      *     listings.
      */
    virtual void derived_print(const display &value) = 0;

    /**
      * The derived_feed method is used to implement the FD opcode,
      * in the cases where it is valid.
      */
    virtual void derived_feed(void) = 0;

    /**
      * The derived_off method is used to implement the OFF key.
      */
    virtual void derived_off(void) = 0;

    /**
      * The derived_update_display method is used when interpreting
      * opcodes to indicate that the display needs to be updated to
      * track the calculator's internal state.
      *
      * @param value
      *     The value to be displayed by the calculator display.
      */
    virtual void derived_update_display(const display &value) = 0;

    /**
      * The printer_timer_start_running method is called by the
      * calculator to inform the derived class that it should kick off a
      * printer delay 420ms timer.  This will happen when opcode_print
      * or opcode_feed opcode is seen.  Derived classes should not
      * assume when printer delays stop and start, instend they will be
      * explicitly informed via this method.
      */
    virtual void printer_timer_start_running(void) = 0;

    /**
      * The accumulate_mode_get method may be used to obtain the state
      * of the #accumulate_mode instance variable.
      */
    bool accumulate_mode_get(void) const;

    /**
      * The mstate_t type is used to describe the state machine
      * used to digit keys and arithmetic keys.  The state machine
      * implementation can be found in the #on_opcode_ope method.
      */
    enum mstate_t
    {
        /**
          * This is the state the calculator is in when it is reset.
          * It is also the state entered after a [+=] or [-=] key.  As
          * such, this state does <b>not</b> imply that #x_register or
          * #y_register are zero (they could be, but don't assume they are).
          */
        mstate_reset,
        mstate_lhs_digits_before_dot,
        mstate_lhs_digits_after_dot,
        mstate_lhs_complete,
        mstate_mul,
        mstate_mul_digits_before_dot,
        mstate_mul_digits_after_dot,
        mstate_mul_complete,
    };

    static const char *mstate_name(mstate_t s);

    mstate_t mstate;

    /**
      * The precro_front_panel instance variable is used to remember the
      * setting selected by the user for the number of decimal places.
      */
    precision_and_rounding precro_front_panel;

    /**
      * The precro_current instance variable is used to remember the
      * number of decimal places and rounding.  This is usually the
      * same as #precro_front_panel instance variable, except after the
      * opcode_round_up, opcode_round_off and opcode_round_down opcodes.
      */
    precision_and_rounding precro_current;

    number x_register;

    number y_register;

    number memory[16];

    /**
      * The program_selector instance variable is used to remember the
      * addressing mode in use.  This controls how the program space is
      * segmented.
      *
      * "Program Selector Switch
      *
      * "When storing a program, executing a program calculation, and checking
      * or clearing a program, program ranges are selected by this switch.
      *
      * 1) Program storage
      *
      *    (i)   When two independent programs are stored at the same time, a
      *          program within 120 steps is stored on side I and a program
      *          within 120 steps is stored on side II.
      *
      *    (ii)  When one program of over 120 steps is stored, the switch is
      *          set in the I+II position.
      *
      *    (iii) When the respective programs at I and II positions exceed
      *          120 steps and 120 steps respectiveIy, the excess is stored
      *          automatically in II or I.
      *
      * 2) Execution of a program
      *
      *    The program is also classified by the changeover among I + II, I and
      *    II in the case of program calculation execution in accordance with
      *    the stored program.
      *
      *    (i)   To execute a program by using the contents of program A, set
      *          the switch in I + II or I position.
      *
      *    (ii)  To execute a program by using the contents of program B, set
      *          the switch in II position.
      *
      *    (iii) When the two programs are stored and they are used as
      *          required, execute program A with the switch set to I position,
      *          and execute program B with the switch set in II position.
      *
      * 3) Check of program contents
      *
      *    Checking of a particular stored program is also accomplished by
      *    selecting I+II, I or II.
      *
      *    (i)   When checking program A, set the selector to either I+II or I.
      *
      *    (ii)  When checking program B, set the selector to II.
      *
      *    (iii) When checking a dual program, set the selector at I for A and
      *          II for B.
      *
      * 4) Clearing contents of program
      *
      *    When clearing a program, the range to be cleared is determined by
      *    the changeover among I+II, I and II.
      *
      *    (i)   To clear I+II range, steps 1 through 240, set the switch in
      *          I+II position.
      *
      *    (ii)  To clear I range, steps 1 through 120, set the switch in I
      *          position.
      *
      *    (iii) To clear II range, steps 121 through 240, set the switch in II
      *          position.
      */
    program_selector_t program_selector;

    /**
      * The address instance variable is used to remember where the
      * next opcode to be accessed is, relative to the start of the
      * #programme array.
      *
      * The 1614P used an antiquated notion of address space, being from
      * 1..240 (rathe rthen a more moder zero-based approach).  By using
      * an array larger than that, we can accommodate this, without
      * constant resort to subracting one.
      */
    unsigned char address;

    /**
      * The programme instance variable is used to remember the store
      * programme of the calculator.  The array is acrually larger
      * then the original ,machine, but than means we don't have to
      * constantly subtract one all the time.
      */
    unsigned char programme[256];

    /**
      * The programme_location instance variable is used to remember the
      * source file location of each of the opcodes in the #programme
      * instance variable.
      */
    location::pointer programme_location[SIZEOF(programme)];

    /**
      * The program_mode instance variable is used to remeber the
      * calculator's current program mode.
      */
    program_mode_t program_mode;

    interpose::pointer interpose_by_program_mode[4];

    enum deferred_action_t
    {
        deferred_action_none,
        deferred_action_multiplication,
        deferred_action_division,
    };

    /**
      * The deferred_action instance variable is used to remember
      * the previous arithmetic operation key.
      */
    deferred_action_t deferred_action;

    /**
      * The update_display_lrn method is used to re-draw the display
      * when in LRN mode.
      */
    void update_display_lrn(void);

    /**
      * The calculate_display_lrn method is used to calculate what
      * would be displayed and/or printed for the given opcodes.
      *
      * @param disp
      *     The retiurned constructed display value.
      * @param addr
      *     The address of the opcode (1..240)
      * @param op
      *     The opcode of interest (0..127)
      */
    void calculate_display_lrn(display &disp, int addr, int op);

    /**
      * The update_display_ope method is used to re-draw the display
      * when in OPE mode.
      */
    void update_display_ope(void);

    enum clock_tick_mode_t
    {
        clock_tick_mode_disabled,
        clock_tick_mode_running,
        clock_tick_mode_entry,
        clock_tick_mode_sj,
        clock_tick_mode_ej,
    };

    /**
      * The clock_tick_mode instance vareiable is used to remember
      * whether or not the calculator thinks the clock is running.  This
      * will be true when in Program Mode OPE, and the user has pressed
      * the Start key (or UJ n).
      */
    clock_tick_mode_t clock_tick_mode;

    /**
      * The printer_delay_active method instance variable is used to
      * remember whether or not a printer delay is active.  This is
      * used to pause instruction execution until the printer finishes
      * printing a line.
      */
    bool printer_delay_active;

    /**
      * The next_opcode method is used to fetch the next opcode from
      * #programme and then advance the #address by one, wrapping as
      * required.
      */
    unsigned char next_opcode(void);

    /**
      * The find_flag_jump method is used to scan the address space,
      * looking for the given flag label (i.e. opcode_fj followed by
      * this value).
      *
      * @param label
      *     The flag jump label to look for
      * @returns
      *     The address of the opcode following the flag jump, or the
      *     lowest address in the address space if not found.
      */
    unsigned find_flag_jump(unsigned char label);

    /**
      * The wrap_address method is used to calculate the appropriate
      * wrapped value for this given address, taking into account
      * #program_selector value.
      *
      * @param addr
      *     The address to be checks, and wrapped if necessary.
      */
    unsigned wrap_address(unsigned addr) const;

    /**
      * The update_display_lamps method is used to consistently set the
      * appropriate lamps for the user interface.
      */
    void update_display_lamps(display &value);

    /**
      * The execute_one_instruction method is used by #clock_tick_event
      * and #on_opcode_ope_che to execute the next instruction.
      */
    void execute_one_instruction(void);

    /**
      * The program_print method is used to print a program listing on
      * the printer.
      */
    void program_print(void);

    /**
      * The accumulate_mode instance variable is used to remember the
      * state of the AM1 key.
      *
      * See the Instruction Manual, p. 9
      *
      * "[AM1] Accumulative Memory 1 Key
      *
      * "For automatic accumulation.  Used in sums and differences of products
      * and quotients.  By depressing the [+=] key or [-=] key, the displayed
      * data is added to or subtracted from Memory 1.  This is a lock key for
      * Memory 1.  (Not used in programming.)"
      */
    bool accumulate_mode;

    /**
      * The call_stack instance variable is used to remember the
      * returned addresses for called subroutines.  It is not clear
      * if the 1614P was even capable of more than one level of calls
      * (although it may explain the last 5 bytes in the #programme
      * space).
      */
    std::list<unsigned> call_stack;

    /**
      * The find_subroutine_flag_jump method is used to scan the address
      * space, looking for the given subroutine flag label (i.e.
      * opcode_sfj followed by this value).
      *
      * @param label
      *     The subroutine flag jump label to look for
      * @returns
      *     The address of the opcode following the subroutine flag
      *     jump, or the lowest address in the address space if not
      *     found.
      */
    unsigned find_subroutine_flag_jump(unsigned char label);

    /**
      * The next_location method is used to obtain the location
      * corresponding to the next #programme location to be executaed
      * (see #address).  The return value could be NULL, meaning no
      * information is available.
      */
    location::pointer next_location(void) const;

    /**
      * The constant_mode instance variable is used to remember the
      * state of the [K] key, and adjust multiplication and division
      * accordingly.
      *
      * Instruction Manual, p. 10
      *
      * "This key is for constant multiplication and division.  When it
      * is depressed, it locks and function state is attained.
      *
      * "Internal calculation is performed in the normal procedure.
      * Thereafter constant multiplication and division can be made by
      * operating the numeric and [+=] keys.
      *
      * "In the case of multiplication, the multiplicant is a constant,
      * and in the case of division, the divisor is a constant.
      *
      * "This key is also used in involutions.  The "n"th power of a
      * figure is obtain by depressing the [+=] key (n-1) times.  Thus,
      * a [x] [+=] [+=] [+=] [+=] gives a<sup>5</sup>.
      */
    bool constant_mode;

    /**
      * The constant_mode_action instance variable is used to remember
      * the appropriate operation (multiplication or division) to apply
      * when the [+=] or [-=] key is used when in constanht mode (when
      * the [K] key is locked down).
      */
    deferred_action_t constant_mode_action;

    /**
      * The constant_mode_rhs instance variable is used to remember the
      * right hand side of the repeated multiplication or division to be
      * performed when the [+=] or [-=] key is pressed when in constant
      * mode (when the [K] key is locked down).
      */
    number constant_mode_rhs;

    /**
      * The constant_mode_eval method is used to evaluate a += or -=
      * key, when the [K] constant key is down.
      *
      * @param neg
      *     false if +=, true if -=
      */
    void constant_mode_eval(bool neg);

    /**
      * The non_constant_mode_eval method is used to evaluate a += or -=
      * key, when the [K] constant key is NOT down.
      *
      * @param neg
      *     false if +=, true if -=
      */
    void non_constant_mode_eval(bool neg);

    /**
      * The eval method is used to evaluate a += or -= key press.
      *
      * @param neg
      *     false if +=, true if -=
      */
    void eval(bool neg);

    /**
      * The eval_mul method is used to evaluate a multiplication or
      * division, in response to a += or -= key press.
      *
      * @param neg
      *     false if +=, true if -=
      */
    void eval_mul(bool neg);

    typedef std::map<std::string, std::string> properties_t;

    /**
      * The properties instance variable is used to remember the
      * properties of the program card set.
      */
    properties_t properties;

    /**
      * The get_property method is used to obtain a specific named
      * property from the set of properties.
      *
      * @param name
      *     The name of the property
      * @param dflt
      *     The value ot be used if the property is not available.
      */
    std::string get_property(const std::string &name,
        const std::string &dflt = "") const;

    /**
      * The get_bool_property method is used to obtain a specific named
      * boolean property from the set of properties.
      *
      * @param name
      *     The name of the property
      * @param dflt
      *     The value ot be used if the property is not available.
      */
    bool get_bool_property(const std::string &name, bool dflt = false) const;

    /**
      * The have_property method is used to determine whether or not the
      * named property is available.
      *
      * @param name
      *     The name of the property
      */
    bool have_property(const std::string &name) const;

    /**
      * The maybe_accumulate method is used to implement the [AM1] mode
      * behaviour.  Addition or subtraction are automagically added to
      * M1 on [+=] or [-=]. However, multiplication and division do NOT
      * change M1 on [+=] or [-=].
      */
    void maybe_accumulate(void);

    /**
      * The strict_memory_number_checking instance variable is used to
      * remember whether or not the key following M3~, SM3~, RM3~ and
      * CM3~ must be a numeric key (true) or is permitted to be any key,
      * of which the lower 4 bits will be used.
      */
    bool strict_memory_number_checking;

    /**
      * The number_of_keys_since_ej instance variable is used to
      * remember how many keys have been pressed since the EJ opcode
      * occurred, to decive which way to branch.  It is also used to
      * know whether or not the ENT opcode should also print.
      */
    int number_of_keys_since_ej;

    /**
      * The copy constructor.
      * Do not use.
      */
    calculator(const calculator &);

    /**
      * The assignment operator.
      * Do not use.
      */
    calculator &operator=(const calculator &);
};

// vim: set ts=8 sw=4 et :
#endif // LIB_CALCULATOR_H
