//
// canola - canon canola 1614p emulator
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIB_DEBUGGER_H
#define LIB_DEBUGGER_H

#include <lib/config.h>
#include <boost/shared_ptr.hpp>

#include <lib/location.h>
#include <lib/number.h>
#include <lib/opcode.h>

/**
  * The debugger class is used to represent the state of an object being
  * used to debug the inner workings of the calculator.
  */
class debugger
{
public:
    typedef boost::shared_ptr<debugger> pointer;

    /**
      * The destructor.
      */
    virtual ~debugger();

    /**
      * The display_x_register method is used to update the displayed
      * value of the X register.
      */
    virtual void display_x_register(const number &value) = 0;

    /**
      * The display_y_register method is used to update the displayed
      * value of the Y register.
      */
    virtual void display_y_register(const number &value) = 0;

    /**
      * The display_memory method is used to update the displayed value
      * of the memory specified.
      *
      * @param n
      *     which memory is being updated (1..14)
      * @param value
      *     the value of the specified memory location
      */
    virtual void display_memory(int n, const number &value) = 0;

    /**
      * The display_address method is used to update the displayed value
      * of the address of the next opcode to be executed.
      */
    virtual void display_address(int addr) = 0;

    /**
      * The display_opcode method is used to update the displayed value
      * of the next opcode to be executed.
      */
    virtual void display_opcode(opcode_t op) = 0;

    /**
      * The display_location method is used to update the displayed
      * value of the source code text, highlighting the section that is
      * the token for the next opcode to be executed.
      */
    virtual void display_location(const location::pointer &where) = 0;

    virtual void display_mstate(const char *text) = 0;

protected:
    /**
      * The default constructor.
      * For use by derived classes only.
      */
    debugger();

private:
    /**
      * The copy constructor.
      * Do not use.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    debugger(const debugger &rhs);

    /**
      * The assignment operator.
      * Do not use.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    debugger &operator=(const debugger &rhs);
};

#endif // LIB_DEBUGGER_H
