//
// canola - canon canola 1614p emulator
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <lib/ac/string.h>
#include <libexplain/output.h>

#include <lib/image/format/jpeg.h>
#include <lib/image/format/png.h>
#include <lib/sizeof.h>


struct table_t
{
    const char *name;
    image::pointer (*create)(const std::string &filename);
    image::pointer (*create_blank)(int width, int height);
};


static table_t table[] =
{
    { "jpg", &image_format_jpeg::create, &image_format_jpeg::create_blank },
    { "jpeg", &image_format_jpeg::create, &image_format_jpeg::create_blank },
    { "png", &image_format_png::create, &image_format_png::create_blank },
};


static std::string
extension(const std::string &path)
{
    const char *s = path.c_str();
    const char *slash = strrchr(s, '/');
    const char *begin = (slash ? slash + 1 : s);
    const char *dot = strrchr(begin, '.');
    return (dot ? dot + 1 : "");
}


image::pointer
image::create_from_file(const std::string &filename)
{
    std::string ext = extension(filename);

    for (const table_t *tp = table; tp < ENDOF(table); ++tp)
    {
        if (std::string(tp->name) == ext)
            return tp->create(filename);
    }

    explain_output_error_and_die
    (
        "file extension \"%s\" is not in a recognised image format, try PNG",
        ext.c_str()
    );
    return image_format_png::create(filename);
}


image::pointer
image::create_blank_from_file(const std::string &filename, int width,
    int height)
{
    std::string ext = extension(filename);

    for (const table_t *tp = table; tp < ENDOF(table); ++tp)
    {
        if (std::string(tp->name) == ext)
            return tp->create_blank(width, height);
    }

    explain_output_error_and_die
    (
        "file extension \"%s\" is not in a recognised image format, try PNG",
        ext.c_str()
    );
    return image_format_png::create_blank(width, height);
}


bool
image::known_file_extension(const std::string &ext)
{
    for (const table_t *tp = table; tp < ENDOF(table); ++tp)
    {
        if (std::string(tp->name) == ext)
            return true;
    }
    return false;
}
