//
// canola - canon canola 1614p emulator
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIB_INTERPOSE_H
#define LIB_INTERPOSE_H

#include <lib/ac/stdarg.h>
#include <boost/shared_ptr.hpp>

#include <lib/format_printf.h>
#include <lib/location.h>
#include <lib/opcode.h>

/**
  * The interpose class is used to represent the state of interpreting
  * key strokes, based on the preceeding keystrokes.  This permits
  * state-ful keystroke processing to be flattened into a single
  * state-less stream.
  */
class interpose
{
public:
    typedef boost::shared_ptr<interpose> pointer;

    /**
      * The destructor.
      */
    virtual ~interpose();

    /**
      * The on_opcode method is used to process keys that have a
      * corresponding opcode.
      *
      * @param op
      *     The opcode (usually a key stroke) of interest
      * @param where
      *     The source file location, or NULL if not available.
      */
    virtual void on_opcode(opcode_t op, const location::pointer &where) = 0;

    /**
      * The strict_memory_number_checking_required method is used to determine
      * whether or not the number keys following M3~ keys (et al) must be
      * followed by a digit key (true) or any key (false).  The actual 1614P
      * silently permits any key, and uses only the lower 4 bits.
      */
    virtual bool strict_memory_number_checking_required(void) const = 0;

protected:
    /**
      * The on_error method is used to report errors processing the key stream.
      *
      * @param fmt
      *     The format of the error.  This controls the numer and types
      *     of the following arguments.
      */
    void on_error(const char *fmt, ...)
                                                            FORMAT_PRINTF(2, 3);

public:
    /**
      * The on_error_v method is used to report errors processing the
      * key stream.
      *
      * @param fmt
      *     The format of the error.  This controls the numer and types
      *     of the following arguments.
      * @param ap
      *     The following arguments.
      */
    virtual void on_error_v(const char *fmt, va_list ap)
                                                        FORMAT_PRINTF(2, 0) = 0;

protected:
    /**
      * The default constructor.
      * For use by derived classes only.
      */
    interpose();

private:
    /**
      * The copy constructor.
      * Do not use.
      */
    interpose(const interpose &);

    /**
      * The assignment operator.
      * Do not use.
      */
    interpose &operator=(const interpose &);
};

#endif // LIB_INTERPOSE_H
