//
// canola - canon canola 1614p emulator
// Copyright (C) 2011, 2012 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIB_LEXER_H
#define LIB_LEXER_H

#include <lib/ac/stdio.h>
#include <string>
#include <list>
#include <map>
#include <boost/shared_ptr.hpp>

#include <lib/format_printf.h>
#include <lib/location.h>
#include <lib/search_path.h>

class calculator; // forward
class lexer_functor; // forward

/**
  * The lexer class is used to represent the processing required to
  * read and understand a script file.  This file contains mnem,onics
  * for each of the keys, and passes them to a calculator instance to
  * execute.
  */
class lexer
{
public:
    /**
      * The destructor.
      */
    virtual ~lexer();

    /**
      * The constructor.
      */
    lexer(const std::string &filename, calculator &subject);

    /**
      * The read_and_execute method is used to read the named text
      * file, and decode the opcodes therein, passing them to the given
      * #subject.
      *
      * @returns
      *     True if no errors, false if at least one error.
      */
    bool read_and_execute(void);

    /**
      * The read_binary_and_execute method is used to read the named
      * binary file, and decode the opcodes therein, passing them to the
      * given #subject.
      *
      * @returns
      *     True if no errors, false if at least one error.
      */
    bool read_binary_and_execute(void);

    /**
      * The test_mode_set method is used to force the lexer into "test
      * mode" where you can enter more things to drive the calculator
      * state.  Must be called before the #read_and_execute method to
      * have any effect.
      *
      * @param state
      *     The state of #test_mode to set
      */
    void test_mode_set(bool state = true);

    /**
      * The search_path_append class method is used to append another
      * directory to the list of directories search for include files.
      */
    static void search_path_append(const std::string &dir);

private:
    /**
      * The subject instance variable is used to remember the claculator
      * that is to receive the code stream, and is also used to report
      * errors.
      */
    calculator &subject;

    /**
      * The source class is used to represent a source file.  Because we
      * may have include files, this is used as elements on a push-down
      * stack of sources.
      */
    class source
    {
    public:
        typedef boost::shared_ptr<source> pointer;

        /**
          * The destructor.
          * It is not virtual; thou shalt not derive from this class.
          */
        ~source();

        /**
          * The create class method is sued to create new dynamically
          * allocated instances of this class.
          *
          * @param filename
          *     The name of the file to be opened
          * @param subject
          *     used to report errors.
          */
        static pointer create(const std::string &filename, calculator &subject);

        /**
          * The getch method is used to obtain the next character.
          *
          * @returns
          *     the next charactare (0..255), or EOF at end-of-file
          */
        int getch(void);

        /**
          * The ungetch method may be used to returna value (obtained
          * from #getch) to the input stream.  Only a single level of
          * pushback is supported.
          *
          * @param c
          *     The value to push back (it can even cope with EOF)
          */
        void ungetch(int c);

        /**
          * This is a hack so that we can report correct addresses for
          * binary files.
          */
        void set_line_number(int n);

        /**
          * The error method is used to report errors, including file
          * name and line number.
          */
        void error(const char *fmt, ...)                    FORMAT_PRINTF(2, 3);

        /**
          * The error_v method is used to report errors, including file
          * name and line number.
          */
        void error_v(const char *fmt, va_list)              FORMAT_PRINTF(2, 0);

        location::pointer get_location(void) const;

    private:
        /**
          * The constructor.
          * It is priovate on purpose, use the #create class method instead.
          *
          * @param filename
          *     The name of the file to be opened
          * @param subject
          *     used to report errors.
          */
        source(const std::string &filename, calculator &subject);

        /**
          * The filename instance variable is used to remember the name of the
          * file being parsed.  There are two choices, logical and physical.
          */
        std::string filename;

        /**
          * The subject instance variable is used to remember the
          * calculator instance we use to report errors.
          */
        calculator &subject;

        /**
          * The fp instance variable is used to remember the open FILE
          * pointer, for the file we are reading.  Will be NULL once EOF
          * has been reached.
          */
        FILE *fp;

        /**
          * The line_number instance variable is used to remember what
          * line number we are currently processing, so that errors can
          * include the filename and the line number.
          */
        int line_number;

        /**
          * The prev_was_newline instance variable is used to remember
          * whether or not the previous character was anewline.  This
          * helps us to make sure errors are reported against the
          * appropriate line.
          */
        bool prev_was_newline;

        /**
          * The position instance variable is used to remember the
          * current byte position within the source file.
          */
        long position;

        /**
          * The close method is used to terminate the rest of the input.
          * After this, #getch will always return EOF.
          */
        void close(void);

        /**
          * The pushback instance variable is used to return one or more
          * characters to the input.
          *
          * We need this because stdio only guarantees a single
          * character of push back, and we need several.
          */
        std::list<unsigned char> pushback;
    };

    /**
      * The src instance variable is used to remember the stack of data
      * sources.  Include files are pushed onto this stack.
      */
    std::list<source::pointer> src;

    /**
      * The test_mode instance variable is used to remember whether or
      * not we are operating on test mode.
      */
    bool test_mode;

    /**
      * The close method is used to close all open input files.  All
      * subsequent #getch calls will report EOF.
      */
    void close(void);

    /**
      * The next_is method is used to see if the next character is the
      * character given.  If it is, that character of input is consumed;
      * if not, no input is consumed.
      *
      * @param c
      *     The character of interest.
      */
    bool next_is(char c);

    /**
      * The next_is method is used to see if the next few characters
      * are the string given.  If it is, thise characters of input are
      * consumed; if not, no input is consumed.
      *
      * @param s
      *     The character string of interest.
      */
    bool next_is(const char *s);

    /**
      * The getch method is used to obtain the next character if input,
      * or EOF if all inpout files, including include files, have been
      * exhausted.
      */
    int getch(void);

    /**
      * The ungetc method is used to return a character (previously
      * obtain from the #getch method) to the input.
      *
      * @param c
      *     The character to be returned (it can even cope with EOF).
      */
    void ungetch(int c);

    /**
      * The error method is used to report errors to the subject.
      * It will include file filename and line number.
      *
      * @param fmt
      *     The error format, and it describes the number and types o
      *     the remaining arguments.
      */
    void error(const char *fmt, ...)                        FORMAT_PRINTF(2, 3);

    /**
      * The open_include_file method is used internally to open include
      * files.  They are permitted to be nested.
      *
      * @param filename
      *     The name of the file to be included.
      */
    void open_include_file(const std::string &filename);

    /**
      * The include_search_path class variable is used to remember the
      * list of directories to be searched for include files.
      */
    static search_path include_search_path;

    typedef boost::shared_ptr<lexer_functor> functor_pointer;
    typedef std::map<std::string, functor_pointer> names_t;

    /**
      * The names insrtance variable is used to remember the table of
      * names and corresponding actions, when parsing text files.
      */
    names_t names;

    /**
      * The load_names method is used to populate the #names instance
      * variable, when needed.
      */
    void load_names(void);

    /**
      * The get_location method is used to obtain the present location
      * within the text of the programme.  It has zero width.
      */
    location::pointer get_location(void) const;

    /**
      * The get_location method is used to obtain the present location
      * within the text of the programme.
      *
      * @param start
      *     relative to this start location, to set the width.
      */
    location::pointer get_location(const location::pointer &start) const;

    /**
      * The card_properties_active instance variable is used to
      * remember whether or not we are between "Card-Pocket-Begin" and
      * "Card-Pocket-End" comment lines.
      */
    bool card_properties_active;

    /**
      * The process_comment method is used to process the text of a
      * comment, looking for card properties.
      *
      * @param text
      *     The text body of the comment (without the introducer character).
      */
    void process_comment(const std::string &text);

    /**
      * The ok instance variable is used to remember whether or not an
      * error has ocurred when processing a file.  True means no error,
      * false means at least one error.
      */
    bool ok;

    /**
      * The default constructor.
      * Do not use.
      */
    lexer();

    /**
      * The copy constructor.
      * Do not use.
      */
    lexer(const lexer &);

    /**
      * The assignment operator.
      * Do not use.
      */
    lexer &operator=(const lexer &);
};

#endif // LIB_LEXER_H
