//
// canola - canon canola 1614p emulator
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIB_LOCATION_H
#define LIB_LOCATION_H

#include <lib/ac/stddef.h>
#include <string>
#include <boost/shared_ptr.hpp>

#include <lib/format_printf.h>

/**
  * The location class is used to represent a span of text from a source
  * file.  The line number is included, but the byte range will be more
  * useful in getting a text view widget to track it.
  */
class location
{
public:
    typedef boost::shared_ptr<location> pointer;

    /**
      * The destructor.
      */
    virtual ~location();

    /**
      * The create class method is used to create new dynamically
      * allocated instance of this class.
      *
      * @param filename
      *     The name of the text source file
      * @param line_number
      *     The line number of the text source file
      * @param byte_start
      *     The byte address of the token within the source file.
      * @param byte_length
      *     The byte length of the token in the source file.
      */
    static pointer create(const std::string &filename, int line_number,
        size_t byte_start, size_t byte_length);

    static location::pointer span(const location::pointer &from,
        const location::pointer &to);

    std::string get_filename(void) const { return filename; }
    int get_line_number(void) const { return line_number; }
    size_t get_start(void) const { return byte_start; }
    size_t get_end(void) const { return (byte_start + byte_length); }

    /// For debugging.
    std::string representation(void) const;

private:
    /**
      * The constructor.
      * It is private on purpose, use the #create class method instead.
      *
      * @param filename
      *     The name of the text source file
      * @param line_number
      *     The line number of the text source file
      * @param byte_start
      *     The byte address of the token within the source file.
      * @param byte_length
      *     The byte length of the token in the source file.
      */
    location(const std::string &filename, int line_number, size_t byte_start,
        size_t byte_length);

    /**
      * The filename instance variable is used to remember the name of
      * the text source file.
      */
    std::string filename;

    /**
      * The line_number instance variable is used to remember the line
      * number of the text source file.
      */
    int line_number;

    /**
      * The byte_start instance variable is used to remember the byte
      * address of the token within the source file.
      */
    size_t byte_start;

    /**
      * The byte_length instance variable is used to remember the byte
      * length of the token in the source file.
      */
    size_t byte_length;

    /**
      * The default constructor.
      * Do not use
      */
    location();

    /**
      * The copy constructor.
      * Do not use
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    location(const location &rhs);

    /**
      * The assignment operator.
      * Do not use
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    location &operator=(const location &rhs);
};

#endif // LIB_LOCATION_H
