//
// canola - canon canola 1614p emulator
// Copyright (C) 2011, 2012 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIB_NUMBER_H
#define LIB_NUMBER_H

#include <lib/config.h>
#include <string>

#include <lib/display.h>
#include <lib/number/z.h>


/**
  * The number class is used to represent the values used by the
  * calculator to perform arithmetic.  It is limited to 16 digits, and a
  * decimal point placed somewhere in those 16 digits.  The minus sign
  * is separate.  The overflow bit indicates carry into the nonexistent
  * next digit.
  */
class number
{
public:
    /**
      * The destructor.
      * It is not virtual; thou shalt not derive from this class.
      */
    ~number();

    /**
      * The default constructor.
      * Value is zero.
      */
    number();

    /**
      * The copy constructor.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    number(const number &rhs);

    /**
      * A constructor.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    number(long rhs);

    /**
      * The assignment operator.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    number &operator=(const number &rhs);

    /**
      * The add-in-situ operator may be used to add another number to
      * this number.
      *
      * @param rhs
      *     The right hand side of the addition.
      */
    void operator+=(const number &rhs);

    /**
      * The addition operator may be used to add two numbers
      * together.
      *
      * @param rhs
      *     The right hand side of the addition.
      */
    number operator+(const number &rhs) const;

    /**
      * The subtract-in-situ operator may be used to subtract another
      * number from this number.
      *
      * @param rhs
      *     The right hand side of the subtraction.
      */
    void operator-=(const number &rhs);

    /**
      * The subtraction operator may be used to add subtract one number
      * from another.
      *
      * @param rhs
      *     The right hand side of the subtraction.
      */
    number operator-(const number &rhs) const;

    /**
      * The negative operator may be used to invert the sign of a
      * number.
      */
    number operator-(void) const;

    /**
      * The multiply-in-situ operator may be used to multiply this
      * number by another number.
      *
      * @param rhs
      *     The right hand side of the multiplication.
      */
    void operator*=(const number &rhs);

    /**
      * The miltiplication operator may be used to multiply two numbers
      * together.
      *
      * @param rhs
      *     The right hand side of the multiplication.
      */
    number operator*(const number &rhs) const;

    /**
      * The divide-in-situ operator may be used to divide this
      * number by another number.
      *
      * @param rhs
      *     The right hand side of the division.
      */
    void operator/=(const number &rhs);

    /**
      * The division operator may be used to divide on number by aother.
      *
      * @param rhs
      *     The right hand side of the division.
      */
    number operator/(const number &rhs) const;

    /**
      * The sqrt method may be used to calculate the square root of a
      * number.  The sign is ignored.
      */
    number sqrt(void) const;

    /**
      * The to_long method is used to extract the integer part of the
      * number into a C long value.  If the value would overflow,
      * LONG_MAX or -LONG_MAX will be returned.
      */
    long to_long(void) const;

    /**
      * The round_up method is used to round this value away from zero,
      * to the given number of decimal places.
      *
      * @param preferred_decimal
      *     The desired number of decimal places, provided that the
      *     number will still fit.
      */
    number round_up(int preferred_decimal) const;

    /**
      * The round_off method is used to round this value towards the nearest,
      * to the given number of decimal places.
      *
      * @param preferred_decimal
      *     The desired number of decimal places, provided that the
      *     number will still fit.
      */
    number round_off(int preferred_decimal) const;

    /**
      * The round_down method is used to round this value towards zero,
      * to the given number of decimal places.
      *
      * @param preferred_decimal
      *     The desired number of decimal places, provided that the
      *     number will still fit.
      */
    number round_down(int preferred_decimal) const;

    /**
      * The to_string method may be used to form a textual
      * representation of this number.
      */
    std::string to_string(void) const;

    /**
      * The get_display method may be used to obtain a representation of
      * this number in a form suitable for displaying on the calculator
      * front panel.
      */
    display get_display(void) const;

    /**
      * The right_shift method may be used to shift all of the digits
      * of thenumber one place to the right, discarding the right most
      * digit.  If there are any decimal places, they are reduced by
      * one.  This is done in situ.
      */
    void right_shift(void);

    /**
      * The equality operator may be used to compare number values.
      *
      * @param rhs
      *     The right hand side of the comparison.
      */
    bool operator==(const number &rhs) const;

    /**
      * The inequality operator may be used to compare number values.
      *
      * @param rhs
      *     The right hand side of the comparison.
      */
    bool operator!=(const number &rhs) const;

    /**
      * The less-than operator may be used to compare number values.
      *
      * @param rhs
      *     The right hand side of the comparison.
      */
    bool operator<(const number &rhs) const;

    /**
      * The less-than_or-equal operator may be used to compare number values.
      *
      * @param rhs
      *     The right hand side of the comparison.
      */
    bool operator<=(const number &rhs) const;

    /**
      * The larger-than operator may be used to compare number values.
      *
      * @param rhs
      *     The right hand side of the comparison.
      */
    bool operator>(const number &rhs) const;

    /**
      * The larger-than_or-equal operator may be used to compare number values.
      *
      * @param rhs
      *     The right hand side of the comparison.
      */
    bool operator>=(const number &rhs) const;

    /**
      * The sign method may be used to obtain the sign of the numkber.
      *
      * @returns
      *     -1 if the number is negative, or 1 if the number is zero or one.
      */
    int sign(void) const;

    /**
      * The fabs method may be used to obtain the the sign-less value of
      * the number.  That is, (n < 0 ? -n : n).
      */
    number fabs(void) const;

    /**
      * The insert_digit_before_dot method is used when building a
      * number, to insert a digit to the left of the decimal point.
      */
    void insert_digit_before_dot(int n);

    /**
      * The insert_digit_after_dot method is used when building a
      * number, to insert a digit to the right of the decimal point.
      */
    void insert_digit_after_dot(int n);

    /**
      * The insert_dot method is used to process a redundant not when
      * constructing a number.
      */
    void insert_dot(void);

    /**
      * The has_overflowed method may be used to determine whether
      * or not this numeric value is too large to fit on the 1614P's
      * display.
      */
    bool has_overflowed(void) const;

    /**
      * The infinity class member may be used to create a value that has
      * overflowed, for use when division by zero occurs (and probably
      * some other cases I haven't found yet).
      */
    static number infinity(void);

private:
    /**
      * The value instance variable is used to remember the value of the
      * 16 digits.  You could use an unsigned long long instead, but we
      * want to produce some of the 1614P's niche behaviour.
      */
    number_z value;

    /**
      * The decimal instance variable is used to remember the number fo
      * decimal places to give the value, if it were to be converted to
      * floating point.
      */
    unsigned decimal;

    /**
      * The negative instance variable is used to remember whether or
      * not the value is negative.
      */
    bool negative;

    static unsigned max(unsigned a, unsigned b) { return (a > b ? a : b); }

    static int max(int a, int b) { return (a > b ? a : b); }

    /**
      * The clear method is used to reset the number back to zero.
      */
    void clear(void);

    /**
      * The is_zero method is used to determine whether or not this
      * number is exactly zero.
      */
    bool is_zero(void) const;

    /**
      * The print_me method is used to print a representation of the
      * number on stderr.  This is for use when debugging.
      *
      * @param file
      *     The source file, use the __FILE__ macro.
      * @param line
      *     The source line, use the __LINE__ macro.
      */
    void print_me(const char *file, int line) const;

    /**
      * A constructor.
      *
      * @param value
      *     The underlying integer value.
      * @param decimal
      *     The number of decimal places.
      * @param negative
      *     Whether or not the number is negative.
      */
    number(const number_z &value, unsigned decimal, bool negative = false);

    /**
      * The positive_minus_positive class method is used to subtract two
      * positive numbers.  This is a helper function for the #operator+
      * and #operator- methods.
      */
    static number positive_minus_positive(const number &lhs, const number &rhs);

    /**
      * The positive_plus_positive class method is used to add two
      * positive numbers.  This is a helper function for the #operator+
      * and #operator- methods.
      */
    static number positive_plus_positive(const number &lhs, const number &rhs);

    /**
      * The truncate_floating_operation class method is used to adjust
      * the number of decimal places of an answer, so that it will fit
      * on the 1614P's display.
      */
    static void truncate_floating_operation(number_z &t3, unsigned &t3ndec);

    /**
      * The strip_redundant_decimal_places class method is used to
      * remove zero digits on the right, if they are to the right of the
      * decimal point.
      */
    static void strip_redundant_decimal_places(number_z &t3, unsigned &t3ndec);
};

// vim: set ts=8 sw=4 et :
#endif // LIB_NUMBER_H
