//
// canola - canon canola 1614p emulator
// Copyright (C) 2012 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIB_NUMBER_Z_H
#define LIB_NUMBER_Z_H

#include <lib/ac/stddef.h>
#include <string>

/**
  * The number_z class is used to represent an arbitrary precision
  * non-negative number, represented as one BCD digit per byte.
  */
class number_z
{
public:
    /**
      * The destructor.
      * It is not virtual, thou shalt not derive from this class.
      */
    ~number_z();

    /**
      * The default constructor.
      * The value will be zero.
      */
    number_z();

    /**
      * A constructor.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    number_z(unsigned long long rhs);

    /**
      * A constructor.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    number_z(unsigned long rhs);

    /**
      * A constructor.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    number_z(unsigned rhs);

    /**
      * The copy constructor.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    number_z(const number_z &rhs);

    /**
      * The assignment operator.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    number_z &operator=(const number_z &rhs);

    void clear(void);

    bool is_zero(void) const;

    bool is_not_zero(void) const { return !is_zero(); }

    bool is_one(void) const;

    number_z &operator+=(const number_z &rhs);

    number_z operator+(const number_z &rhs) const;

    number_z &operator-=(const number_z &rhs);

    number_z operator-(const number_z &rhs) const;

    number_z &operator*=(const number_z &rhs);

    number_z operator*(const number_z &rhs) const;

    number_z &operator/=(const number_z &rhs);

    number_z operator/(const number_z &rhs) const;

    number_z &operator%=(const number_z &rhs);

    number_z operator%(const number_z &rhs) const;

    /**
      * The shift_left method is used to shift the number the given
      * number of decimal digits to the left.  Zeros will be shifted
      * into the vacated positions.
      *
      * @param ndigits
      *     The number of decimal places to shift by.
      *
      * @note
      *     We do not overload the << operator, because it is for shifting
      *     binary bits, whereas this method shifts by decimal places.
      */
    number_z shift_left(size_t ndigits) const;

    /**
      * The shift_right method is used to shift the number the given
      * number of decimal digits to the right.  Digits on the right are
      * lost.
      *
      * @param ndigits
      *     The number of decimal places to shift by.
      *
      * @note
      *     We do not overload the >> operator, because it is for shifting
      *     binary bits, whereas this method shifts by decimal places.
      */
    number_z shift_right(size_t ndigits) const;

    bool operator<(const number_z &rhs) const;

    bool operator<=(const number_z &rhs) const;

    bool operator>(const number_z &rhs) const;

    bool operator>=(const number_z &rhs) const;

    bool operator==(const number_z &rhs) const;

    bool operator!=(const number_z &rhs) const;

    /**
      * The get method is used to get the value of the digit, specified
      * by an index.
      *
      * @param digit_number
      *     The nunber of the digit, starting at 0 for the right hand
      *     digit and working rightwards as the digit_number increases.
      * @returns
      *     The value of the digit (0..9), or 0 if you are requesting a
      *     digit beyond the most significant digit available.  This is
      *     <b>not</b> in ASCII.
      */
    unsigned get(size_t digit_number) const;

    size_t get_digit_count(void) const { return digits_used; }

    /**
      * The right_16 method is used to contruct a new numer from this
      * one, using only the right-most 16 digits, and ignoring any
      * further to the left.
      */
    number_z right_16(void) const;

    std::string representation(void) const;

    /**
      * The one class method is used to obtain a constant 1 value, for
      * use in expressions.  This avoid the ambiguity warnings from the
      * compiler.
      */
    static number_z one(void);

    void print_me(const char *file, int line, const char *caption) const;

    /**
      * The build method is used when building number from left to
      * right, e.g. in a lexical analyzer.  The number is shifted
      * left on decimal place, and the given digits is made the least
      * significant decimal digit.
      *
      * @param c
      *     The digit to be added on the right.
      */
    void build(unsigned c);

    /**
      * The to_long method is used to obtaim a C long representation of
      * the value.  If it is too big, LONG_MAX will be returned.
      */
    long to_long(void) const;

private:
    /**
      * The digits_used instance variable is used to remember
      * the number of used positions in the #digits array.
      *
      * assert(digits_used <= digits_allocated);
      */
    size_t digits_used;

    /**
      * The digits_allocated instance variable is used to remember the
      * number of digits that have been dynamically allocated for the
      * #digits array.
      */
    size_t digits_allocated;

    typedef unsigned char digit_t;

    /**
      * The digits instance variable is used to remember the base
      * address of a dynamically allocated array of digits.  It grows
      * exponentially, so that #set is an O(1) operation.
      *
      * assert(!digits == !digits_allocated);
      *
      * The digits between #digits_used and #digits_allocated are
      * explicitly set to zero.
      */
    digit_t *digits;

    /**
      * The is_value method is used to determine whether or not the
      * internal representation of the BCD integer is self-consistent.
      */
    bool is_valid(void) const;

    /**
      * The set method is used to set a specific decimal digit to a
      * specific decimal value.
      *
      * @param dignum
      *     The digit number; 0 is on the rights, more significant
      *     digits have larger numbers.
      * @param digval
      *     The value of the digits, always in the range 0..9
      */
    void set(size_t dignum, unsigned digval);

    /**
      * The div_mod method is used by the #operator/ and #operator%
      * methods to perform the calculation.
      *
      * assert(quotient * denominator + remainder == numerator);
      *
      * @param numerator
      *     The number above the line in the division.
      * @param denominator
      *     The number below the line in the division.
      * @param quotient
      *     The whole number of times the denominator can be subtracted
      *     from the numerator.
      * @param remainder
      *     What is left over after the numerator is done.
      */
    static bool div_mod(const number_z &numerator, const number_z &denominator,
        number_z &quotient, number_z &remainder);

    /**
      * The compare class method is used to determine the relationship
      * between two integers.  This helper method is used to the
      * relational operators ( #operator<, #operator<=, #operator>,
      * #operator>=, #operator== and #operator!=).
      *
      * @param lhs
      *     The left hand side of the comparison.
      * @param rhs
      *     The right hand side of the comparison.
      * @returns
      *     <0 if lhs is less than rhs,
      *      0 if lhs is equal to rhs
      *     >0 if lhs is greather than rhs
      */
    static int compare(const number_z &lhs, const number_z &rhs);
};

// vim: set ts=8 sw=4 et :
#endif // LIB_NUMBER_Z_H
