//
// canola - canon canola 1614p emulator
// Copyright (C) 2012 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIB_POLYFIT_H
#define LIB_POLYFIT_H

/**
  * The polyfit class is used to represent the state of a least squares
  * polynomial curve fit.
  */
class polyfit
{
public:
    /**
      * The destructor.
      */
    virtual ~polyfit();

    /**
      * The constructor.
      *
      * @param order
      *     The desired order of polynomial; 1 is linear, 2 is quadratic, etc.
      */
    polyfit(int order = 1);

    /**
      * The enter method is used to add another point to the scatter
      * plot being solved to fit a polynomial.
      *
      * @param x
      *     The X coordinate of the point
      * @param y
      *     The Y coordinate of the point
      */
    void enter(double x, double y);

    enum solve_t
    {
        solve_ok,
        solve_singular,
        solve_insufficient
    };

    /**
      * The solve method is used to solve the point set to produce a
      * polynomial, once all of the points have been entered using the
      * #enter method.
      */
    solve_t solve(void);

    /**
      * The coefficient method is used to obtain the nth coefficient
      * of the polynomial.  Will only produce valid results after the
      * #solve method has been called.
      *
      * @param n
      *     The number of the coefficient of interest,
      *     corresponding to the x**n term.
      */
    double coefficient(int n);

    /**
      * The evaluate method is used to evaluate the fitted polynomial,
      * at the given coordinate.
      *
      * @param x
      *     The X coordinate of interest.
      * @returns
      *     The corresponding Y coordinate.
      */
    double evaluate(double x);

    /**
      * The print method is used to print the polynomial to the standard
      * output.  Mostly only useful for debugging.
      */
    void print(void) const;

    /**
      * The enough_points method is used to determine whether or not
      * sufficient points are available for a meaningful solution.
      */
    bool enough_points(void) const { return matrix[0][0] >= rows; }

private:
    /**
      * The row instance variable is used to remember the number of rows
      * in the data #matrix.
      */
    unsigned int rows;

    /**
      * The cols instance variable is used to remember the number of
      * columns in the data #matrix.
      */
    unsigned int cols;

    typedef long double data_t;

    /**
      * The coefficients instance variable is used to remember the
      * polynomial coefficients, set bu the #solve method.
      */
    data_t *coefficients;

    /**
      * The matrix instance variable is used to remember the various
      * summs of paired data, use by the #solve method to fit the
      * polynomial to the data.
      */
    data_t **matrix;

    /**
      * The epsilon class variable is used to remember the value below
      * which a value is considered to be zero.
      */
    static data_t epsilon;

    /**
      * The mat_alloc method is used to allocate a matrix, filled with
      * zero, that is #rows rows higth and #cols columns wide.  This is
      * used to allocate the #matrix instance variable, and also the
      * temporary working space used by the #solve method.
      */
    data_t **mat_alloc();

    /**
      * The mat_dump method is used to print a prepresentation of the
      * matrix to the standard output.  Only usefil for debugging.
      *
      * @param data
      *     The matrix to be printed.
      */
    void mat_dump(data_t **data);

    /**
      * The mat_free method is used to release the memory consumed by a
      * matrix.
      *
      * @param data
      *     The matrix to be released.
      */
    void mat_free(data_t **data);

    /**
      * The default constructor.  Do not use.
      */
    polyfit();

    /**
      * The copy constructor.  Do not use.
      */
    polyfit(const polyfit &);

    /**
      * The assignment operator.  Do not use.
      */
    polyfit &operator=(const polyfit &);
};

#endif // LIB_POLYFIT_H
