//
// canola - canon canola 1614p emulator
// Copyright (C) 2011, 2012 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#ifndef LIB_PROGRAM_CARD_FUNCTOR_IMAGE_H
#define LIB_PROGRAM_CARD_FUNCTOR_IMAGE_H

#include <cairomm/context.h>

#include <lib/program_card/functor.h>

/**
  * The program_card_functor class is used to represent the processing
  * required to draw Canola 1614P program card images.
  */
class program_card_functor_image:
    public program_card_functor
{
public:
    typedef boost::shared_ptr<program_card_functor_image> pointer;

    /**
      * The destructor.
      */
    virtual ~program_card_functor_image();

    /**
      * The factory class method is used to create a functor suitable for use
      * drawing images of programs cards.  The file extension of the filename
      * is examplined, for things like ".png" or ".pdf", etc.
      *
      * @param filename
      *     the name of the file to be written.
      * @returns
      *     A managed pointer to a suitable instance.  If it doesn't recognise
      *     the file extension, it issues a fatal error and exits.
      */
    static pointer factory(const std::string &filename);

    // See base class for documentation.
    void set_disassemble(bool value);

protected:
    /**
      * The default constructor.
      * For use by derived classes only.
      */
    program_card_functor_image();

    /**
      * The draw method is used to draw an image of a program card
      * into the given drawing context.  Because Cairo has numerous
      * drawing surfaces, this allows card images to be PNG file, or
      * post script, or Gtk windows, etc.
      *
      * @param cr
      *     The cairo graphics context to draw on.
      * @param data
      *     The program data to be drawn on the card image
      * @param data_size
      *     The length of the program to be drawn.
      */
    void draw(const Cairo::RefPtr<Cairo::Context> &cr,
        const unsigned char *data, size_t data_size);

    /**
      * The draw_envelope method is used to draw an image of a card
      * envelope, for keeping your cards sets in.
      *
      * @param cr
      *     The cairo graphics context to draw on.
      */
    void draw_envelope(const Cairo::RefPtr<Cairo::Context> &cr);

    void set_position(int card_num, int card_count);

    // See base class for documentation.
    void set_title(const std::string &text);

    // See base class for documentation.
    void set_arrow(bool value);

    // See base class for documentation.
    void set_color(const std::string &text);

    // See base class for documentation.
    void set_perforations(bool value);

    // See base class for documentation.
    void set_trim_corners(bool value);

    // See base class for documentation.
    void set_style(const std::string &text);

    // See base class for documentation.
    virtual void set_sfj_list(const std::string &text);

    // See base class for documentation.
    virtual void set_fj_list(const std::string &text);

    // See base class for documentation.
    virtual void set_memory_list(const std::string &text);

    // See base class for documentation.
    virtual void set_steps(int value);

private:
    /**
      * The trim_corners instance variable is used to remember whether
      * or not the leading corners are to be trimmed.
      */
    bool trim_corners;

    /**
      * The perforations instance variable is used to remember whether
      * or not the bits that are not holes should have perforations
      * drawn in their place.
      */
    bool perforations;

    /**
      * The disassemble instance variable is used to remember whether or
      * not to disassemble the opcode down the right hand side of the
      * card.
      */
    bool disassemble;

    /**
      * The arrow instance variable is used to remember whether or not
      * to draw a directional arrow on the card.
      */
    bool arrow;

    int card_num;

    int card_count;

    std::string color;

    std::string title;

    enum style_t
    {
        style_aus,
        style_ibm,
        style_blank
    };

    style_t style;

    std::string sfj_list;
    unsigned accuracy;
    unsigned step_count;
    std::string fj_list;
    std::string memory_list;

    void outline(const Cairo::RefPtr<Cairo::Context> &cr);

    /**
      * The copy constructor.
      * Do not use.
      *
      * @param rhs
      *     The right hand side of the initialization.
      */
    program_card_functor_image(const program_card_functor_image &rhs);

    /**
      * The assignment operator.
      * Do not use.
      *
      * @param rhs
      *     The right hand side of the assignment.
      */
    program_card_functor_image &operator=(
        const program_card_functor_image &rhs);
};

#endif // LIB_PROGRAM_CARD_FUNCTOR_IMAGE_H
