/*
 * canola - canon canola 1614p emulator
 * Copyright (C) 2011, 2012 Peter Miller
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 3, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

%{

#include <lib/ac/stdlib.h>
#include <lib/ac/stdio.h>
#include <libexplain/ferror.h>

#include <lib/number.h>

#include <test_number/grammar.h>
#include <test_number/lex.h>

#define YYDEBUG 1

%}

%token COMMA
%token DIV
%token EOLN
%token EQ
%token FLOOR
%token GE
%token GT
%token JUNK
%token LE
%token LP
%token LT
%token MINUS
%token MUL
%token NE
%token NUMBER
%token PLUS
%token RP
%token SQRT

%left LT LE GT GE EQ NE
%left PLUS MINUS
%left MUL DIV
%right unary

%union {
    number *lv_number;
}

%type <lv_number> expression
%type <lv_number> NUMBER

%{

int yyparse(void);


void
grammar(void)
{
    yyparse();
    lex_end();
}


extern int yydebug;


void
grammar_debugging(void)
{
    yydebug = 1;
}

%}

%%

run
    : /* empty */
    | run line
    ;

line
    : EOLN
    | expression EOLN
        {
            printf("%s\n", $1->to_string().c_str());
            delete $1;
            explain_ferror_or_die(stdout);
        }
    | error EOLN
    ;

expression
    : NUMBER
        { $$ = $1; }
    | LP expression RP
        { $$ = $2; }
    | SQRT LP expression RP
        {
            number x = *$3;
            delete $3;
            $$ = new number(x.sqrt());
        }
    | FLOOR LP expression COMMA expression RP
        {
            number x = *$3;
            delete $3;
            number n = *$5;
            delete $5;
            if (n < 0)
                n = 0;
            else if (n > 15)
                n = 15;
            $$ = new number(x.round_down(n.to_long()));
        }
    | PLUS expression
        %prec unary
        { $$ = $2; }
    | MINUS expression
        %prec unary
        {
            $$ = new number(-*$2);
            delete $2;
        }
    | expression MUL expression
        {
            number lhs(*$1);
            delete $1;
            number rhs(*$3);
            delete $3;
            number result = lhs * rhs;
            $$ = new number(result);
        }
    | expression DIV expression
        {
            number lhs(*$1);
            delete $1;
            number rhs(*$3);
            delete $3;
            number result = lhs / rhs;
            $$ = new number(result);
        }
    | expression PLUS expression
        {
            number lhs(*$1);
            delete $1;
            number rhs(*$3);
            delete $3;
            number result = lhs + rhs;
            $$ = new number(result);
        }
    | expression MINUS expression
        {
            number lhs(*$1);
            delete $1;
            number rhs(*$3);
            delete $3;
            number result = lhs - rhs;
            $$ = new number(result);
        }
    | expression LT expression
        {
            number lhs(*$1);
            delete $1;
            number rhs(*$3);
            delete $3;
            bool result = (lhs < rhs);
            $$ = new number(result ? 1 : 0);
        }
    | expression LE expression
        {
            number lhs(*$1);
            delete $1;
            number rhs(*$3);
            delete $3;
            bool result = (lhs <= rhs);
            $$ = new number(result ? 1 : 0);
        }
    | expression GT expression
        {
            number lhs(*$1);
            delete $1;
            number rhs(*$3);
            delete $3;
            bool result = (lhs > rhs);
            $$ = new number(result ? 1 : 0);
        }
    | expression GE expression
        {
            number lhs(*$1);
            delete $1;
            number rhs(*$3);
            delete $3;
            bool result = (lhs >= rhs);
            $$ = new number(result ? 1 : 0);
        }
    | expression EQ expression
        {
            number lhs(*$1);
            delete $1;
            number rhs(*$3);
            delete $3;
            bool result = (lhs == rhs);
            $$ = new number(result ? 1 : 0);
        }
    | expression NE expression
        {
            number lhs(*$1);
            delete $1;
            number rhs(*$3);
            delete $3;
            bool result = (lhs != rhs);
            $$ = new number(result ? 1 : 0);
        }
    ;


/* vim: set ts=8 sw=4 et : */
