//
// canola - canon canola 1614p emulator
// Copyright (C) 2011, 2012 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <lib/ac/stdio.h>
#include <libexplain/getchar.h>
#include <libexplain/output.h>
#include <libexplain/ungetc.h>

#include <lib/number.h>

#include <test_number/lex.h>
#include <test_number/grammar.yacc.h>


static int number_of_errors;
static int line_number = 1;
static bool prev_was_eoln;


void
lex_end(void)
{
    if (number_of_errors > 0)
    {
        explain_output_error_and_die
        (
            "found %d error%s",
            number_of_errors,
            (number_of_errors == 1 ? "" : "s")
        );
    }
}


static bool
have(char c)
{
    int t = explain_getchar_or_die();
    if (t == EOF)
        return false;
    if (t == c)
        return true;
    explain_ungetc_or_die(t, stdin);
    return false;
}


int
grammar_lex(void)
{
    if (prev_was_eoln)
        ++line_number;
    for (;;)
    {
        int c = explain_getchar_or_die();
        switch (c)
        {
        case EOF:
            return 0;

        case '\t':
            break;

        case '\n':
            prev_was_eoln = true;
            return EOLN;

        case ' ':
            break;

        case '!':
            if (have('='))
                return NE;
            return JUNK;

        case '(':
            return LP;

        case ')':
            return RP;

        case '*':
            return MUL;

        case '+':
            return PLUS;

        case ',':
            return COMMA;

        case '-':
            return MINUS;

        case '/':
            return DIV;

        case '0': case '1': case '2': case '3': case '4':
        case '5': case '6': case '7': case '8': case '9':
            {
                number n;
                bool dotted = false;
                for (;;)
                {
                    if (dotted)
                        n.insert_digit_after_dot(c - '0');
                    else
                        n.insert_digit_before_dot(c - '0');
                    get_more:
                    c = explain_getchar_or_die();
                    switch (c)
                    {
                    case EOF:
                        break;

                    case '.':
                        dotted = true;
                        n.insert_dot();
                        goto get_more;

                    case '0': case '1': case '2': case '3': case '4':
                    case '5': case '6': case '7': case '8': case '9':
                        continue;

                    default:
                        explain_ungetc_or_die(c, stdin);
                        break;
                    }
                    break;
                }
                grammar_lval.lv_number = new number(n);
            }
            return NUMBER;

        case '<':
            if (have('='))
                return LE;
            return LT;

        case '=':
            if (have('='))
                return EQ;
            return JUNK;

        case '>':
            if (have('='))
                return GE;
            return GT;

        case '_':
        case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
        case 'G': case 'H': case 'I': case 'J': case 'K': case 'L':
        case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
        case 'S': case 'T': case 'U': case 'V': case 'W': case 'X':
        case 'Y': case 'Z':
        case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
        case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':
        case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
        case 's': case 't': case 'u': case 'v': case 'w': case 'x':
        case 'y': case 'z':
            {
                std::string name;
                for (;;)
                {
                    name += (char)c;

                    c = explain_getchar_or_die();
                    switch (c)
                    {
                    case EOF:
                        break;

                    case '0': case '1': case '2': case '3': case '4':
                    case '5': case '6': case '7': case '8': case '9':
                    case '_':
                    case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
                    case 'G': case 'H': case 'I': case 'J': case 'K': case 'L':
                    case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
                    case 'S': case 'T': case 'U': case 'V': case 'W': case 'X':
                    case 'Y': case 'Z':
                    case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
                    case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':
                    case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
                    case 's': case 't': case 'u': case 'v': case 'w': case 'x':
                    case 'y': case 'z':
                        continue;

                    default:
                        explain_ungetc_or_die(c, stdin);
                        break;
                    }
                    break;
                }
                if (name == "sqrt")
                    return SQRT;
                if (name == "round_down")
                    return FLOOR;
            }
            return JUNK;

        default:
            return JUNK;
        }
    }
}


void
grammar_error(const char *text)
{
    explain_output_error("%d: %s", line_number, text);
    ++number_of_errors;
}


// vim: set ts=8 sw=4 et :
