//
// canola - canon canola 1614p emulator
// Copyright (C) 2012 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <lib/ac/stdio.h>
#include <libexplain/getchar.h>
#include <libexplain/output.h>
#include <libexplain/ungetc.h>

#include <lib/number/z.h>

#include <test_number_z/lex.h>
#include <test_number_z/grammar.yacc.h>


static int number_of_errors;
static int line_number = 1;
static bool prev_was_eoln;


void
lex_end(void)
{
    if (number_of_errors > 0)
    {
        explain_output_error_and_die
        (
            "found %d error%s",
            number_of_errors,
            (number_of_errors == 1 ? "" : "s")
        );
    }
}


static bool
have(char c)
{
    int t = explain_getchar_or_die();
    if (t == EOF)
        return false;
    if (t != c)
    {
        ungetc(t, stdin);
        return false;
    }
    return true;
}


int
grammar_lex(void)
{
    if (prev_was_eoln)
        ++line_number;
    for (;;)
    {
        int c = explain_getchar_or_die();
        switch (c)
        {
        case EOF:
            return 0;

        case '\t':
            break;

        case '\n':
            prev_was_eoln = true;
            return EOLN;

        case ' ':
            break;

        case '!':
            if (have('='))
                return NE;
            return JUNK;

        case '%':
            return MOD;

        case '(':
            return LP;

        case ')':
            return RP;

        case '*':
            return MUL;

        case '+':
            return PLUS;

        case '-':
            return MINUS;

        case '/':
            return DIV;

        case '0': case '1': case '2': case '3': case '4':
        case '5': case '6': case '7': case '8': case '9':
            {
                number_z n;
                for (;;)
                {
                    n.build(c - '0');

                    c = explain_getchar_or_die();
                    switch (c)
                    {
                    case EOF:
                        break;

                    case '0': case '1': case '2': case '3': case '4':
                    case '5': case '6': case '7': case '8': case '9':
                        continue;

                    default:
                        explain_ungetc_or_die(c, stdin);
                        break;
                    }
                    break;
                }
                grammar_lval.lv_number = new number_z(n);
            }
            return NUMBER;

        case '<':
            if (have('='))
                return LE;
            return LT;

        case '=':
            if (have('='))
                return EQ;
            return JUNK;

        case '>':
            if (have('='))
                return GE;
            return GT;

        default:
            return JUNK;
        }
    }
}


void
grammar_error(const char *text)
{
    explain_output_error("%d: %s", line_number, text);
    ++number_of_errors;
}



// vim: set ts=8 sw=4 et :
