//
// canola - canon canola 1614p emulator
// Copyright (C) 2011 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License, version 3, as
// published by the Free Software Foundation.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this program. If not, see <http://www.gnu.org/licenses/>.
//

#include <lib/ac/ctype.h>
#include <lib/ac/stdio.h>
#include <lib/ac/stdlib.h>
#include <lib/ac/string.h>
#include <lib/ac/unistd.h>
#include <libexplain/fclose.h>
#include <libexplain/fflush.h>
#include <libexplain/fgets.h>
#include <libexplain/fopen.h>
#include <libexplain/output.h>
#include <libexplain/program_name.h>
#include <list>
#include <string>

#include <lib/print_version.h>


static void
usage(void)
{
    const char *prog = explain_program_name_get();
    fprintf(stderr, "Usage: %s <filename>...\n", prog);
    fprintf(stderr, "       %s -V\n", prog);
    exit(1);
}


static void
grope(const char *filename)
{
    FILE *fp = explain_fopen_or_die(filename, "r");

    {
        const char *cp = filename;
        while (cp[0] == 'b' && cp[1] == 'l')
            cp += 2;
        if (*cp == '/')
            filename = cp + 1;
    }

    typedef std::list<std::string> lines_t;
    lines_t lines;

    for (;;)
    {
        char line[2000];
        if (!explain_fgets_or_die(line, sizeof(line), fp))
            break;
        size_t len = strlen(line);
        while (len > 0 && isspace((unsigned char)line[len - 1]))
            --len;
        lines.push_back(std::string(line, len));
    }
    explain_fclose_or_die(fp);

    lines_t::const_iterator it = lines.begin();
    int linum = 1;
    while (it != lines.end())
    {
        const char *p = strstr(it->c_str(), "Ques" "tion:");
        if (p)
        {
            p += 9;
            while (*p && isspace((unsigned char)*p))
                ++p;
            if (*p)
            {
                printf("<li>\n");
                printf("%s\n", p);
                printf("<br/>\n");
                printf("<small>(file %s, line %d)</small>\n", filename, linum);
            }
        }
        else
        {
            p = strstr(it->c_str(), "<ques" "tion>");
            if (p)
            {
                printf("<li>%s\n", p + 10);

                lines_t::const_iterator ep = it;
                ++ep;
                for (;;)
                {
                    if (ep == lines.end())
                    {
                        explain_output_error_and_die
                        (
                            "%s: %d: unterminated question",
                            filename,
                            linum
                        );
                    }
                    if (strstr(ep->c_str(), "</ques" "tion>"))
                        break;
                    const char *s = ep->c_str();
                    while (isspace((unsigned char)*s))
                        ++s;
                    while (ispunct((unsigned char)*s))
                        ++s;
                    while (isspace((unsigned char)*s))
                        ++s;
                    if (*s)
                        printf("%s\n", s);
                    ++ep;
                }

                printf("<br/>\n");
                printf("<small>(file %s, line %d)</small>\n", filename, linum);
            }
        }
        ++it;
        ++linum;
    }
}


int
main(int argc, char **argv)
{
    for (;;)
    {
        int c = getopt(argc, argv, "V");
        if (c == EOF)
            break;
        switch (c)
        {
        case 'V':
            print_version();
            return 0;

        default:
            usage();
        }
    }
    if (optind == argc)
        usage();

    printf("<html>\n");
    printf("<body>\n");
    printf("<h1 align=\"center\">\n");
    printf("<img src=\"canola.128.png\" align=\"left\" alt=\"\" />\n");
    printf("<img src=\"canola.128.png\" align=\"right\" alt=\"\" />\n");
    printf("<a href=\"index.html\" >Canola</a> Questions </h1>\n");
    printf("Questions to be put to a real 1614P.\n");
    printf("<p>\n");
    printf("All these questions are with the decimal point dial\n");
    printf("set to &ldquo;F&rdquo;, and the rounding switch set to\n");
    printf("&ldquo;5/4&rdquo;.\n");
    printf("<p>\n");
    printf("The [SM3] notation means the [SM3~] key followed by the [3]\n");
    printf("key, similarly for memories larger than 3, and similarly for\n");
    printf("[+M<i>n</i>], [SM<i>n</i>], [CM<i>n</i>], and [RM<i>n</i>]\n");
    printf("notations.  Think of it as the opcode, rather than the key\n");
    printf("sequence needed to produce the opcode.\n");
    printf("<p>\n");
    printf("<ol>\n");
    for (;;)
    {
        grope(argv[optind++]);
        if (optind >= argc)
            break;
    }
    printf("</ol>\n");
    printf("</body>\n");
    printf("</html>\n");

    explain_fflush_or_die(stdout);

    return 0;
}
